//cpu8040.c
//cpu emulation, real implementation

#include <stdio.h>
#include "cpu8040.h"

using namespace xatremu;

void CPU8040::CheckCPUSpecials(void) {
	//if serial int set then trigger int in cpu
	if(m_ExternalIntEnabled && m_IntPossible && !m_Serial->StateIntLine()) {
#ifdef TESTSTATE
	printf("Int triggered\n");
#endif
		m_internalStack[r_sp] = r_PC & 0xff;
		++r_sp;
		UpdatePSW();
		m_internalStack[r_sp] = ((r_PC & 0xf00) >> 8) + (r_psw & 0xf0);
		++r_sp;
		r_sp &= 0xf;
		r_PC = 3;
		m_IntPossible = false;
		return;
	};
	if(m_totalCycles >= m_NextSpecialEvent) {
		CheckCPUSpecialEvents();
	}
};

void CPU8040::CheckCPUSpecialEvents() {
	//if Timer increase trigger it
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTTIMERINCREASE]) {
		m_SpecialEventTable[CPUEVENTTIMERINCREASE] = MAXCYCLE;
		if (m_TRunning) {
			++r_T;
			m_SpecialEventTable[CPUEVENTTIMERINCREASE] = m_totalCycles + 480;
			if (r_T == 0) {
				r_TF = true;
				if (m_TCNTIntEnabled && m_IntPossible) {
					//trigger int
					m_internalStack[r_sp] = r_PC & 0xff;
					++r_sp;
					UpdatePSW();
					m_internalStack[r_sp] = ((r_PC & 0xf00) >> 8) + (r_psw & 0xf0);
					++r_sp;
					r_sp &= 0xf;
					r_PC = 7;
					m_IntPossible = false;
					DetermineNextEventCount();
					return;
				};
			};
		};
		DetermineNextEventCount();
	};

	//if serial input bits are coming trigger serial
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTSERIALIN]) {
		m_SpecialEventTable[CPUEVENTSERIALIN] = MAXCYCLE;
		m_Serial->T0NextBit();
		DetermineNextEventCount();
	};

	//if serial output started trigger output
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTSERIALOUT]) {
		m_SpecialEventTable[CPUEVENTSERIALOUT] = MAXCYCLE;
		m_Serial->P17NextBit();
		DetermineNextEventCount();

	};
	//check motor on
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTWD1772MOTOR]) {
		m_SpecialEventTable[CPUEVENTWD1772MOTOR] = MAXCYCLE;
		m_WD1772->CheckMotor();
		DetermineNextEventCount();
	};
	//if head step delay is due
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTWD1772HEADSTEP]) {
		m_SpecialEventTable[CPUEVENTWD1772HEADSTEP] = MAXCYCLE;
		m_WD1772->CheckHeadStepDelay();
		DetermineNextEventCount();
	};
	//if head settle delay is due
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTWD1772HEADCALM]) {
                m_SpecialEventTable[CPUEVENTWD1772HEADCALM] = MAXCYCLE;
                m_WD1772->CheckHeadCalmDelay();
                DetermineNextEventCount();
	};
	//if next data from/to WD1772 call callback
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTWD1772NEXTDATA]) {
		m_SpecialEventTable[CPUEVENTWD1772NEXTDATA] = MAXCYCLE;
		m_WD1772->CheckDRQEvent();
		DetermineNextEventCount();
	};
	//if search for right sector timed out
	if(m_totalCycles >= m_SpecialEventTable[CPUEVENTWD1772DATATIMEOUT]) {
		m_SpecialEventTable[CPUEVENTWD1772DATATIMEOUT] = MAXCYCLE;
		m_WD1772->DataSearchTimeOut();
		DetermineNextEventCount();
	};
};

//determine next event, so that not all events must be checked every cycle
void CPU8040::DetermineNextEventCount(void) {
	int64_t tmp = MAXCYCLE;
	for (int loop = 0; loop < SPECIALEVENTCOUNT; ++loop) {
		if (m_SpecialEventTable[loop] < tmp)
			tmp = m_SpecialEventTable[loop];
	};
	m_NextSpecialEvent = tmp;
};

void CPU8040::SetNextSpecialEvent(unsigned char eventType, int64_t addCycles) {
	if (eventType >= SPECIALEVENTCOUNT) throw;
	if (addCycles == MAXCYCLE) {
		m_SpecialEventTable[eventType] = MAXCYCLE;
	}
	else {
		m_SpecialEventTable[eventType] = m_totalCycles + addCycles;
	};
	DetermineNextEventCount();
};

//execute next instruction in program memory
void CPU8040::ExecuteInstruction(void) {
	if (r_PC < m_ROMLen) {
		(this->* CPUFuncTable[GetInstrMem(r_PC)])();

	};
};

void CPU8040::RunCPU(){
	if(m_delayCycles > 0){
		--m_delayCycles;
		++m_totalCycles;
		return;
	};
	CheckCPUSpecials();
#ifdef TESTSTATE
//	if (((r_PC < 0x100) && (r_PC > 0x180)) || ((r_PC & 0xf00) == 0x800) || ((r_PC & 0xf00) == 0x900))
//		printf("r_PC: %x\n", r_PC);
//	if (((r_PC & 0xf00) == 0x400) || (((r_PC & 0xf00) == 0x200) && (r_PC) >= 0x2b5) || ((r_PC & 0xf00) == 0x300))
//	if (((r_PC & 0xf00) == 0x400) || (((r_PC & 0xf00) == 0x000) && (r_PC) >= 0x2a) || ((r_PC & 0xf00) == 0x300))
//		printf("r_PC: %x\n", r_PC);
//	static int trackpc = 0;
//	if (((r_PC & 0xf00) == 0x200) && (r_PC != 0x227)){
//		printf("r_PC: %x\n", r_PC);
//		trackpc = 100;
//	}
//	else if (trackpc > 0){
//		printf("r_PC: %x\n", r_PC);
//		--trackpc;
//	};
#endif
	ExecuteInstruction();
};

void CPU8040::DelayCPU(int64_t myNoOfCycles){
	m_delayCycles += myNoOfCycles;
};

int64_t CPU8040::GetCPUCycle(){
	return m_totalCycles;
};

void CPU8040::RunCPUForCycles(unsigned int numOfCycles){
	m_storedCycles += numOfCycles;
	while(m_totalCycles < m_storedCycles)
		RunCPU();
};

CPU8040::CPU8040() {
	m_WD1772 = NULL;
	m_Serial = NULL;
	m_instrMemPtr = NULL;
	m_DriveNumber = 0;

//00-0f
CPUFuncTable[0] = &CPU8040::I_NOP;
CPUFuncTable[1] = &CPU8040::I_UNDEF;
CPUFuncTable[2] = &CPU8040::I_OUTL_BUS_A;
CPUFuncTable[3] = &CPU8040::I_ADD_A_IMM;
CPUFuncTable[4] = &CPU8040::I_JMP_CODE;
CPUFuncTable[5] = &CPU8040::I_EN_I;
CPUFuncTable[6] = &CPU8040::I_UNDEF;
CPUFuncTable[7] = &CPU8040::I_DEC_A;
CPUFuncTable[8] = &CPU8040::I_INS_A_BUS;
CPUFuncTable[9] = &CPU8040::I_IN_A_P1;
CPUFuncTable[0xa] = &CPU8040::I_IN_A_P2;
CPUFuncTable[0xb] = &CPU8040::I_UNDEF;
CPUFuncTable[0xc] = &CPU8040::I_MOVD_A_P4;
CPUFuncTable[0xd] = &CPU8040::I_MOVD_A_P5;
CPUFuncTable[0xe] = &CPU8040::I_MOVD_A_P6;
CPUFuncTable[0xf] = &CPU8040::I_MOVD_A_P7;

//10-1f
CPUFuncTable[0x10] = &CPU8040::I_INC_AT_R0;
CPUFuncTable[0x11] = &CPU8040::I_INC_AT_R1;
CPUFuncTable[0x12] = &CPU8040::I_JB0_CODE;
CPUFuncTable[0x13] = &CPU8040::I_ADDC_A_IMM;
CPUFuncTable[0x14] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0x15] = &CPU8040::I_DIS_I;
CPUFuncTable[0x16] = &CPU8040::I_JTF_CODE;
CPUFuncTable[0x17] = &CPU8040::I_INC_A;
CPUFuncTable[0x18] = &CPU8040::I_INC_R0;
CPUFuncTable[0x19] = &CPU8040::I_INC_R1;
CPUFuncTable[0x1a] = &CPU8040::I_INC_R2;
CPUFuncTable[0x1b] = &CPU8040::I_INC_R3;
CPUFuncTable[0x1c] = &CPU8040::I_INC_R4;
CPUFuncTable[0x1d] = &CPU8040::I_INC_R5;
CPUFuncTable[0x1e] = &CPU8040::I_INC_R6;
CPUFuncTable[0x1f] = &CPU8040::I_INC_R7;

//20-2f
CPUFuncTable[0x20] = &CPU8040::I_XCH_A_AT_R0;
CPUFuncTable[0x21] = &CPU8040::I_XCH_A_AT_R1;
CPUFuncTable[0x22] = &CPU8040::I_UNDEF;
CPUFuncTable[0x23] = &CPU8040::I_MOV_A_IMM;
CPUFuncTable[0x24] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0x25] = &CPU8040::I_EN_TCNTI;
CPUFuncTable[0x26] = &CPU8040::I_JNT0_CODE;
CPUFuncTable[0x27] = &CPU8040::I_CLR_A;
CPUFuncTable[0x28] = &CPU8040::I_XCH_A_R0;
CPUFuncTable[0x29] = &CPU8040::I_XCH_A_R1;
CPUFuncTable[0x2a] = &CPU8040::I_XCH_A_R2;
CPUFuncTable[0x2b] = &CPU8040::I_XCH_A_R3;
CPUFuncTable[0x2c] = &CPU8040::I_XCH_A_R4;
CPUFuncTable[0x2d] = &CPU8040::I_XCH_A_R5;
CPUFuncTable[0x2e] = &CPU8040::I_XCH_A_R6;
CPUFuncTable[0x2f] = &CPU8040::I_XCH_A_R7;

//30-3f
CPUFuncTable[0x30] = &CPU8040::I_XCHD_A_AT_R0;
CPUFuncTable[0x31] = &CPU8040::I_XCHD_A_AT_R1;
CPUFuncTable[0x32] = &CPU8040::I_JB1_CODE;
CPUFuncTable[0x33] = &CPU8040::I_UNDEF;
CPUFuncTable[0x34] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0x35] = &CPU8040::I_DIS_TCNTI;
CPUFuncTable[0x36] = &CPU8040::I_JT0_CODE;
CPUFuncTable[0x37] = &CPU8040::I_CPL_A;
CPUFuncTable[0x38] = &CPU8040::I_UNDEF;
CPUFuncTable[0x39] = &CPU8040::I_OUTL_P1_A;
CPUFuncTable[0x3a] = &CPU8040::I_OUTL_P2_A;
CPUFuncTable[0x3b] = &CPU8040::I_UNDEF;
CPUFuncTable[0x3c] = &CPU8040::I_MOVD_P4_A;
CPUFuncTable[0x3d] = &CPU8040::I_MOVD_P5_A;
CPUFuncTable[0x3e] = &CPU8040::I_MOVD_P6_A;
CPUFuncTable[0x3f] = &CPU8040::I_MOVD_P7_A;

//40-4f
CPUFuncTable[0x40] = &CPU8040::I_ORL_A_AT_R0;
CPUFuncTable[0x41] = &CPU8040::I_ORL_A_AT_R1;
CPUFuncTable[0x42] = &CPU8040::I_MOV_A_T;
CPUFuncTable[0x43] = &CPU8040::I_ORL_A_IMM;
CPUFuncTable[0x44] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0x45] = &CPU8040::I_STRT_CNT;
CPUFuncTable[0x46] = &CPU8040::I_JNT1_CODE;
CPUFuncTable[0x47] = &CPU8040::I_SWAP_A;
CPUFuncTable[0x48] = &CPU8040::I_ORL_A_R0;
CPUFuncTable[0x49] = &CPU8040::I_ORL_A_R1;
CPUFuncTable[0x4a] = &CPU8040::I_ORL_A_R2;
CPUFuncTable[0x4b] = &CPU8040::I_ORL_A_R3;
CPUFuncTable[0x4c] = &CPU8040::I_ORL_A_R4;
CPUFuncTable[0x4d] = &CPU8040::I_ORL_A_R5;
CPUFuncTable[0x4e] = &CPU8040::I_ORL_A_R6;
CPUFuncTable[0x4f] = &CPU8040::I_ORL_A_R7;

//50-5f
CPUFuncTable[0x50] = &CPU8040::I_ANL_A_AT_R0;
CPUFuncTable[0x51] = &CPU8040::I_ANL_A_AT_R1;
CPUFuncTable[0x52] = &CPU8040::I_JB2_CODE;
CPUFuncTable[0x53] = &CPU8040::I_ANL_A_IMM;
CPUFuncTable[0x54] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0x55] = &CPU8040::I_STRT_T;
CPUFuncTable[0x56] = &CPU8040::I_JT1_CODE;
CPUFuncTable[0x57] = &CPU8040::I_DA_A;
CPUFuncTable[0x58] = &CPU8040::I_ANL_A_R0;
CPUFuncTable[0x59] = &CPU8040::I_ANL_A_R1;
CPUFuncTable[0x5a] = &CPU8040::I_ANL_A_R2;
CPUFuncTable[0x5b] = &CPU8040::I_ANL_A_R3;
CPUFuncTable[0x5c] = &CPU8040::I_ANL_A_R4;
CPUFuncTable[0x5d] = &CPU8040::I_ANL_A_R5;
CPUFuncTable[0x5e] = &CPU8040::I_ANL_A_R6;
CPUFuncTable[0x5f] = &CPU8040::I_ANL_A_R7;

//60-6f
CPUFuncTable[0x60] = &CPU8040::I_ADD_A_AT_R0;
CPUFuncTable[0x61] = &CPU8040::I_ADD_A_AT_R1;
CPUFuncTable[0x62] = &CPU8040::I_MOV_T_A;
CPUFuncTable[0x63] = &CPU8040::I_UNDEF;
CPUFuncTable[0x64] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0x65] = &CPU8040::I_STOP_TCNT;
CPUFuncTable[0x66] = &CPU8040::I_UNDEF;
CPUFuncTable[0x67] = &CPU8040::I_RRC_A;
CPUFuncTable[0x68] = &CPU8040::I_ADD_A_R0;
CPUFuncTable[0x69] = &CPU8040::I_ADD_A_R1;
CPUFuncTable[0x6a] = &CPU8040::I_ADD_A_R2;
CPUFuncTable[0x6b] = &CPU8040::I_ADD_A_R3;
CPUFuncTable[0x6c] = &CPU8040::I_ADD_A_R4;
CPUFuncTable[0x6d] = &CPU8040::I_ADD_A_R5;
CPUFuncTable[0x6e] = &CPU8040::I_ADD_A_R6;
CPUFuncTable[0x6f] = &CPU8040::I_ADD_A_R7;

//70-7f
CPUFuncTable[0x70] = &CPU8040::I_ADDC_A_AT_R0;
CPUFuncTable[0x71] = &CPU8040::I_ADDC_A_AT_R1;
CPUFuncTable[0x72] = &CPU8040::I_JB3_CODE;
CPUFuncTable[0x73] = &CPU8040::I_UNDEF;
CPUFuncTable[0x74] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0x75] = &CPU8040::I_ENT0_CLK;
CPUFuncTable[0x76] = &CPU8040::I_JF1_CODE;
CPUFuncTable[0x77] = &CPU8040::I_RR_A;
CPUFuncTable[0x78] = &CPU8040::I_ADDC_A_R0;
CPUFuncTable[0x79] = &CPU8040::I_ADDC_A_R1;
CPUFuncTable[0x7a] = &CPU8040::I_ADDC_A_R2;
CPUFuncTable[0x7b] = &CPU8040::I_ADDC_A_R3;
CPUFuncTable[0x7c] = &CPU8040::I_ADDC_A_R4;
CPUFuncTable[0x7d] = &CPU8040::I_ADDC_A_R5;
CPUFuncTable[0x7e] = &CPU8040::I_ADDC_A_R6;
CPUFuncTable[0x7f] = &CPU8040::I_ADDC_A_R7;

//80-8f
CPUFuncTable[0x80] = &CPU8040::I_MOVX_A_AT_R0;
CPUFuncTable[0x81] = &CPU8040::I_MOVX_A_AT_R1;
CPUFuncTable[0x82] = &CPU8040::I_UNDEF;
CPUFuncTable[0x83] = &CPU8040::I_RET;
CPUFuncTable[0x84] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0x85] = &CPU8040::I_CLR_F0;
CPUFuncTable[0x86] = &CPU8040::I_JNI_CODE;
CPUFuncTable[0x87] = &CPU8040::I_UNDEF;
CPUFuncTable[0x88] = &CPU8040::I_ORL_BUS_IMM;
CPUFuncTable[0x89] = &CPU8040::I_ORL_P1_IMM;
CPUFuncTable[0x8a] = &CPU8040::I_ORL_P2_IMM;
CPUFuncTable[0x8b] = &CPU8040::I_UNDEF;
CPUFuncTable[0x8c] = &CPU8040::I_ORLD_P4_A;
CPUFuncTable[0x8d] = &CPU8040::I_ORLD_P5_A;
CPUFuncTable[0x8e] = &CPU8040::I_ORLD_P6_A;
CPUFuncTable[0x8f] = &CPU8040::I_ORLD_P7_A;

//90-9f
CPUFuncTable[0x90] = &CPU8040::I_MOVX_AT_R0_A;
CPUFuncTable[0x91] = &CPU8040::I_MOVX_AT_R1_A;
CPUFuncTable[0x92] = &CPU8040::I_JB4_CODE;
CPUFuncTable[0x93] = &CPU8040::I_RETR;
CPUFuncTable[0x94] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0x95] = &CPU8040::I_CPL_F0;
CPUFuncTable[0x96] = &CPU8040::I_JNZ_CODE;
CPUFuncTable[0x97] = &CPU8040::I_CLR_C;
CPUFuncTable[0x98] = &CPU8040::I_ANL_BUS_IMM;
CPUFuncTable[0x99] = &CPU8040::I_ANL_P1_IMM;
CPUFuncTable[0x9a] = &CPU8040::I_ANL_P2_IMM;
CPUFuncTable[0x9b] = &CPU8040::I_UNDEF;
CPUFuncTable[0x9c] = &CPU8040::I_ANLD_P4_A;
CPUFuncTable[0x9d] = &CPU8040::I_ANLD_P5_A;
CPUFuncTable[0x9e] = &CPU8040::I_ANLD_P6_A;
CPUFuncTable[0x9f] = &CPU8040::I_ANLD_P7_A;

//a0-af
CPUFuncTable[0xa0] = &CPU8040::I_MOV_AT_R0_A;
CPUFuncTable[0xa1] = &CPU8040::I_MOV_AT_R1_A;
CPUFuncTable[0xa2] = &CPU8040::I_UNDEF;
CPUFuncTable[0xa3] = &CPU8040::I_MOVP_A_AT_A;
CPUFuncTable[0xa4] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0xa5] = &CPU8040::I_CLR_F1;
CPUFuncTable[0xa6] = &CPU8040::I_UNDEF;
CPUFuncTable[0xa7] = &CPU8040::I_CPL_C;
CPUFuncTable[0xa8] = &CPU8040::I_MOV_R0_A;
CPUFuncTable[0xa9] = &CPU8040::I_MOV_R1_A;
CPUFuncTable[0xaa] = &CPU8040::I_MOV_R2_A;
CPUFuncTable[0xab] = &CPU8040::I_MOV_R3_A;
CPUFuncTable[0xac] = &CPU8040::I_MOV_R4_A;
CPUFuncTable[0xad] = &CPU8040::I_MOV_R5_A;
CPUFuncTable[0xae] = &CPU8040::I_MOV_R6_A;
CPUFuncTable[0xaf] = &CPU8040::I_MOV_R7_A;

//b0-bf
CPUFuncTable[0xb0] = &CPU8040::I_MOV_AT_R0_IMM;
CPUFuncTable[0xb1] = &CPU8040::I_MOV_AT_R1_IMM;
CPUFuncTable[0xb2] = &CPU8040::I_JB5_CODE;
CPUFuncTable[0xb3] = &CPU8040::I_JMPP_AT_A;
CPUFuncTable[0xb4] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0xb5] = &CPU8040::I_CPL_F1;
CPUFuncTable[0xb6] = &CPU8040::I_JF0_CODE;
CPUFuncTable[0xb7] = &CPU8040::I_UNDEF;
CPUFuncTable[0xb8] = &CPU8040::I_MOV_R0_IMM;
CPUFuncTable[0xb9] = &CPU8040::I_MOV_R1_IMM;
CPUFuncTable[0xba] = &CPU8040::I_MOV_R2_IMM;
CPUFuncTable[0xbb] = &CPU8040::I_MOV_R3_IMM;
CPUFuncTable[0xbc] = &CPU8040::I_MOV_R4_IMM;
CPUFuncTable[0xbd] = &CPU8040::I_MOV_R5_IMM;
CPUFuncTable[0xbe] = &CPU8040::I_MOV_R6_IMM;
CPUFuncTable[0xbf] = &CPU8040::I_MOV_R7_IMM;

//c0-cf
CPUFuncTable[0xc0] = &CPU8040::I_UNDEF;
CPUFuncTable[0xc1] = &CPU8040::I_UNDEF;
CPUFuncTable[0xc2] = &CPU8040::I_UNDEF;
CPUFuncTable[0xc3] = &CPU8040::I_UNDEF;
CPUFuncTable[0xc4] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0xc5] = &CPU8040::I_SEL_RB0;
CPUFuncTable[0xc6] = &CPU8040::I_JZ_CODE;
CPUFuncTable[0xc7] = &CPU8040::I_MOV_A_PSW;
CPUFuncTable[0xc8] = &CPU8040::I_DEC_R0;
CPUFuncTable[0xc9] = &CPU8040::I_DEC_R1;
CPUFuncTable[0xca] = &CPU8040::I_DEC_R2;
CPUFuncTable[0xcb] = &CPU8040::I_DEC_R3;
CPUFuncTable[0xcc] = &CPU8040::I_DEC_R4;
CPUFuncTable[0xcd] = &CPU8040::I_DEC_R5;
CPUFuncTable[0xce] = &CPU8040::I_DEC_R6;
CPUFuncTable[0xcf] = &CPU8040::I_DEC_R7;

//d0-df
CPUFuncTable[0xd0] = &CPU8040::I_XRL_A_AT_R0;
CPUFuncTable[0xd1] = &CPU8040::I_XRL_A_AT_R1;
CPUFuncTable[0xd2] = &CPU8040::I_JB6_CODE;
CPUFuncTable[0xd3] = &CPU8040::I_XRL_A_IMM;
CPUFuncTable[0xd4] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0xd5] = &CPU8040::I_SEL_RB1;
CPUFuncTable[0xd6] = &CPU8040::I_UNDEF;
CPUFuncTable[0xd7] = &CPU8040::I_MOV_PSW_A;
CPUFuncTable[0xd8] = &CPU8040::I_XRL_A_R0;
CPUFuncTable[0xd9] = &CPU8040::I_XRL_A_R1;
CPUFuncTable[0xda] = &CPU8040::I_XRL_A_R2;
CPUFuncTable[0xdb] = &CPU8040::I_XRL_A_R3;
CPUFuncTable[0xdc] = &CPU8040::I_XRL_A_R4;
CPUFuncTable[0xdd] = &CPU8040::I_XRL_A_R5;
CPUFuncTable[0xde] = &CPU8040::I_XRL_A_R6;
CPUFuncTable[0xdf] = &CPU8040::I_XRL_A_R7;

//e0-ef
CPUFuncTable[0xe0] = &CPU8040::I_UNDEF;
CPUFuncTable[0xe1] = &CPU8040::I_UNDEF;
CPUFuncTable[0xe2] = &CPU8040::I_UNDEF;
CPUFuncTable[0xe3] = &CPU8040::I_MOVP3_A_AT_A;
CPUFuncTable[0xe4] = &CPU8040::I_JMP_CODE;
CPUFuncTable[0xe5] = &CPU8040::I_SEL_MB0;
CPUFuncTable[0xe6] = &CPU8040::I_JNC_CODE;
CPUFuncTable[0xe7] = &CPU8040::I_RL_A;
CPUFuncTable[0xe8] = &CPU8040::I_DJNZ_R0_CODE;
CPUFuncTable[0xe9] = &CPU8040::I_DJNZ_R1_CODE;
CPUFuncTable[0xea] = &CPU8040::I_DJNZ_R2_CODE;
CPUFuncTable[0xeb] = &CPU8040::I_DJNZ_R3_CODE;
CPUFuncTable[0xec] = &CPU8040::I_DJNZ_R4_CODE;
CPUFuncTable[0xed] = &CPU8040::I_DJNZ_R5_CODE;
CPUFuncTable[0xee] = &CPU8040::I_DJNZ_R6_CODE;
CPUFuncTable[0xef] = &CPU8040::I_DJNZ_R7_CODE;

//f0-ff
CPUFuncTable[0xf0] = &CPU8040::I_MOV_A_AT_R0;
CPUFuncTable[0xf1] = &CPU8040::I_MOV_A_AT_R1;
CPUFuncTable[0xf2] = &CPU8040::I_JB7_CODE;
CPUFuncTable[0xf3] = &CPU8040::I_UNDEF;
CPUFuncTable[0xf4] = &CPU8040::I_CALL_CODE;
CPUFuncTable[0xf5] = &CPU8040::I_SEL_MB1;
CPUFuncTable[0xf6] = &CPU8040::I_JC_CODE;
CPUFuncTable[0xf7] = &CPU8040::I_RLC_A;
CPUFuncTable[0xf8] = &CPU8040::I_MOV_A_R0;
CPUFuncTable[0xf9] = &CPU8040::I_MOV_A_R1;
CPUFuncTable[0xfa] = &CPU8040::I_MOV_A_R2;
CPUFuncTable[0xfb] = &CPU8040::I_MOV_A_R3;
CPUFuncTable[0xfc] = &CPU8040::I_MOV_A_R4;
CPUFuncTable[0xfd] = &CPU8040::I_MOV_A_R5;
CPUFuncTable[0xfe] = &CPU8040::I_MOV_A_R6;
CPUFuncTable[0xff] = &CPU8040::I_MOV_A_R7;

};

CPU8040::~CPU8040() {
};

void CPU8040::I_ADD_A_AT_R0(void) {
	unsigned short int tmp = r_A;
	tmp += m_internalRAM[r_R0];
	r_C = tmp & 0xff00;
	r_A = tmp & 0xff;	
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_ADD_A_AT_R1(void) {
        unsigned short int tmp = r_A;
        tmp += m_internalRAM[r_R1];
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_IMM(void) {
	unsigned short int tmp = r_A;
	++r_PC;
	tmp += GetInstrMem(r_PC);
	r_C = tmp & 0xff00;
	r_A = tmp & 0xff;
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_ADD_A_R0(void) {
	unsigned short int tmp = r_A + r_R0;
	r_C = tmp & 0xff00;
	r_A = tmp & 0xff;
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_ADD_A_R1(void) {
        unsigned short int tmp = r_A + r_R1;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_R2(void) {
        unsigned short int tmp = r_A + r_R2;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_R3(void) {
        unsigned short int tmp = r_A + r_R3;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_R4(void) {
        unsigned short int tmp = r_A + r_R4;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_R5(void) {
        unsigned short int tmp = r_A + r_R5;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_R6(void) {
        unsigned short int tmp = r_A + r_R6;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADD_A_R7(void) {
        unsigned short int tmp = r_A + r_R7;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_AT_R0(void) {
        unsigned short int tmp = r_A + r_C;
        tmp += m_internalRAM[r_R0];
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_AT_R1(void) {
        unsigned short int tmp = r_A + r_C;
        tmp += m_internalRAM[r_R1];
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_IMM(void) {
       	unsigned short int tmp = r_A + r_C;
	++r_PC;
	tmp += GetInstrMem(r_PC);
	r_C = tmp & 0xff00;
	r_A = tmp & 0xff;
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_ADDC_A_R0(void) {
        unsigned short int tmp = r_A + r_R0 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R1(void) {
        unsigned short int tmp = r_A + r_R1 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R2(void) {
        unsigned short int tmp = r_A + r_R2 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R3(void) {
        unsigned short int tmp = r_A + r_R3 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R4(void) {
        unsigned short int tmp = r_A + r_R4 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R5(void) {
        unsigned short int tmp = r_A + r_R5 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R6(void) {
        unsigned short int tmp = r_A + r_R6 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ADDC_A_R7(void) {
        unsigned short int tmp = r_A + r_R7 + r_C;
        r_C = tmp & 0xff00;
        r_A = tmp & 0xff;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_AT_R0(void) {
        r_A &= m_internalRAM[r_R0];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_AT_R1(void) {
        r_A &= m_internalRAM[r_R1];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_IMM(void) {
	++r_PC;
	r_A &= GetInstrMem(r_PC);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_ANL_A_R0(void) {
        r_A &= r_R0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R1(void) {
        r_A &= r_R1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R2(void) {
        r_A &= r_R2;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R3(void) {
        r_A &= r_R3;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R4(void) {
        r_A &= r_R4;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R5(void) {
        r_A &= r_R5;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R6(void) {
        r_A &= r_R6;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_A_R7(void) {
        r_A &= r_R7;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ANL_BUS_IMM(void) {
	throw;
/*        unsigned short int a = 0;
        a = GetInstrMem(r_PC + 1);
        printf("%04x\tANL\tBUS,#%02x\n", r_PC, a);
        r_PC = r_PC + 2;*/
};

void CPU8040::I_ANL_P1_IMM(void) {
	++r_PC;
	unsigned char tmp = GetInstrMem(r_PC);
	tmp &= r_Port1;
	NotifyChangePort1(r_Port1, tmp);
	r_Port1 = tmp;
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_ANL_P2_IMM(void) {
	++r_PC;
        r_Port2 = ((GetInstrMem(r_PC) & r_Port2) & 0x3f) | (m_DriveNumber << 6);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_ANLD_P4_A(void) {
	throw;
        printf("%04x\tANLD\tP4,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_ANLD_P5_A(void) {
	throw;
        printf("%04x\tANLD\tP5,A\n", r_PC);
        ++r_PC;
};      

void CPU8040::I_ANLD_P6_A(void) {
	throw;
        printf("%04x\tANLD\tP6,A\n", r_PC);
        ++r_PC;
};      

void CPU8040::I_ANLD_P7_A(void) {
	throw;
        printf("%04x\tANLD\tP7,A\n", r_PC);
        ++r_PC;
};      

void CPU8040::I_CALL_CODE(void) {
	r_PC += 2;
	m_internalStack[r_sp] = r_PC & 0xff;
	++r_sp;
	UpdatePSW();
	m_internalStack[r_sp] = ((r_PC & 0xf00) >> 8) + (r_psw & 0xf0);
	++r_sp;
	r_sp &= 0xf;
	r_PC = (r_dbf ? 0x800 : 0) + ((GetInstrMem(r_PC - 2) & 0xe0) << 3) + GetInstrMem(r_PC - 1);
        m_totalCycles += 2;
};

void CPU8040::I_CLR_A(void) {
	r_A = 0;
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_CLR_C(void) {
        r_C = false;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_CLR_F0(void) {
        r_F0 = false;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_CLR_F1(void) {
        r_F1 = false;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_CPL_A(void) {
        r_A = ~r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_CPL_C(void) {
        r_C = ~r_C;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_CPL_F0(void) {
        r_F0 = ~r_F0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_CPL_F1(void) {
        r_F1 = ~r_F1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DA_A(void) {
	throw;
        printf("%04x\tDA\tA\n", r_PC);
        ++r_PC;
};

void CPU8040::I_DEC_A(void) {
        --r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R0(void) {
        --r_R0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R1(void) {
        --r_R1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R2(void) {
        --r_R2;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R3(void) {
        --r_R3;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R4(void) {
        --r_R4;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R5(void) {
        --r_R5;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R6(void) {
        --r_R6;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DEC_R7(void) {
        --r_R7;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_DIS_I(void) {
	m_ExternalIntEnabled = false;
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_DIS_TCNTI(void) {
	m_TCNTIntEnabled = false;
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_DJNZ_R0_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
	--r_R0;
	if (r_R0) {
		r_PC = (r_PC & 0xff00) + tmp;
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R1_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
	--r_R1;
	if (r_R1) {
		r_PC = (r_PC & 0xff00) + tmp;
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R2_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
	--r_R2;
	if (r_R2) {
		r_PC = (r_PC & 0xff00) + tmp;
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R3_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
        --r_R3;
        if (r_R3) { 
                r_PC = (r_PC & 0xff00) + tmp; 
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R4_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
        --r_R4;
        if (r_R4) { 
                r_PC = (r_PC & 0xff00) + tmp;
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R5_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
        --r_R5;
        if (r_R5) { 
                r_PC = (r_PC & 0xff00) + tmp;
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R6_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
        --r_R6;
        if (r_R6) { 
                r_PC = (r_PC & 0xff00) + tmp;
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_DJNZ_R7_CODE(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
        --r_R7;
        if (r_R7) { 
                r_PC = (r_PC & 0xff00) + tmp;
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_EN_I(void) {
       	m_ExternalIntEnabled = true;
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_EN_TCNTI(void) {
	m_TCNTIntEnabled = true;
	++r_PC;
	++m_totalCycles;
};      

void CPU8040::I_ENT0_CLK(void) {
	throw;
        printf("%04x\tENT0\tCLK\n", r_PC);
        ++r_PC;
};      

void CPU8040::I_IN_A_P1(void) {
	r_A = (r_Port1 & 0xfb) + (m_WD1772->GetIntRq() ? 0x04 : 0);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_IN_A_P2(void) {
       	r_A = (r_Port2 & 0x3f) + (m_DriveNumber << 6);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_INC_A(void) {
	++r_A;
        ++r_PC;
	++m_totalCycles;
};

void CPU8040::I_INC_AT_R0(void) {
        ++m_internalRAM[r_R0];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_AT_R1(void) {
        ++m_internalRAM[r_R1];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R0(void) {
        ++r_R0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R1(void) {
        ++r_R1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R2(void) {
        ++r_R2;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R3(void) {
        ++r_R3;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R4(void) {
        ++r_R4;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R5(void) {
        ++r_R5;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R6(void) {
        ++r_R6;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INC_R7(void) {
        ++r_R7;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_INS_A_BUS(void) {
	throw;
        printf("%04x\tINS\tA,BUS\n", r_PC);
        ++r_PC;
};

void CPU8040::I_JB0_CODE(void) {
        ++r_PC;
	if (r_A & 0x01) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB1_CODE(void) {
        ++r_PC;
	if (r_A & 0x02) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB2_CODE(void) {
        ++r_PC;
	if (r_A & 0x04) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB3_CODE(void) {
        ++r_PC;
	if (r_A & 0x08) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB4_CODE(void) {
        ++r_PC;
	if (r_A & 0x10) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB5_CODE(void) {
        ++r_PC;
	if (r_A & 0x20) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB6_CODE(void) {
        ++r_PC;
	if (r_A & 0x40) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JB7_CODE(void) {
        ++r_PC;
	if (r_A & 0x80) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JC_CODE(void) {
        ++r_PC;
	if (r_C) {
		r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
	}
	else {
		++r_PC;
	};
        m_totalCycles += 2;
};

void CPU8040::I_JF0_CODE(void) {
        ++r_PC;
        if (r_F0) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JF1_CODE(void) {
        ++r_PC;
        if (r_F1) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        } 
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JMP_CODE(void) {
        r_PC = (r_dbf ? 0x800 : 0) + ((GetInstrMem(r_PC) & 0xe0) << 3) + GetInstrMem(r_PC + 1);
	if(r_PC > 0xe90)
		r_PC &= 0x7ff;
        m_totalCycles += 2;
};

void CPU8040::I_JMPP_AT_A(void) {
        r_PC = (r_PC & 0xf00) + GetInstrMem(r_A);
        m_totalCycles += 2;
};

void CPU8040::I_JNC_CODE(void) {
        ++r_PC;
        if (!r_C) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JNI_CODE(void) {
        ++r_PC;
        if (!m_Serial->StateIntLine()) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JNT0_CODE(void) {
        ++r_PC;
        if (m_Serial->StateT0Line()) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JNT1_CODE(void) {
        ++r_PC;
        if (!m_WD1772->StateT1Line()) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JNZ_CODE(void) {
        ++r_PC;
        if (r_A) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JT0_CODE(void) {
        ++r_PC;
        if (!m_Serial->StateT0Line()) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_JT1_CODE(void) {
        ++r_PC;
        if (m_WD1772->StateT1Line()) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

//one timer step is oscillator clock / 480
void CPU8040::I_JTF_CODE(void) {
        ++r_PC;
        if (r_TF) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
	r_TF = false;
        m_totalCycles += 2;
};

void CPU8040::I_JZ_CODE(void) {
        ++r_PC;
        //if (!r_A || (r_PC == 0xc2)) {
        if (!r_A) {
                r_PC = (r_PC & 0xff00) + GetInstrMem(r_PC);
        }
        else {
                ++r_PC;
        };
        m_totalCycles += 2;
};

void CPU8040::I_MOV_A_IMM(void) {
	++r_PC;
	r_A = GetInstrMem(r_PC);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_MOV_A_AT_R0(void) {
        r_A = m_internalRAM[r_R0];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_AT_R1(void) {
        r_A = m_internalRAM[r_R1];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_PSW(void) {
	UpdatePSW();
	r_A = r_psw;
        ++r_PC;
	++m_totalCycles;
};

void CPU8040::I_MOV_A_R0(void) {
        r_A = r_R0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R1(void) {
        r_A = r_R1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R2(void) {
        r_A = r_R2;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R3(void) {
        r_A = r_R3;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R4(void) {
        r_A = r_R4;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R5(void) {
        r_A = r_R5;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R6(void) {
        r_A = r_R6;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_R7(void) {
        r_A = r_R7;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_A_T(void) {
        r_A = r_T;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_AT_R0_A(void) {
#ifdef TESTSTATE
	printf("%4x MOV @R0,A %2x\n",r_PC, r_A);
#endif
        m_internalRAM[r_R0] = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_AT_R1_A(void) {
        m_internalRAM[r_R1] = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_AT_R0_IMM(void) {
	++r_PC;
        m_internalRAM[r_R0] = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_AT_R1_IMM(void) {
        ++r_PC;
        m_internalRAM[r_R1] = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_PSW_A(void) {
	r_sp = (r_A & 0x7) * 2;
	r_C = r_A & 0x80;
	r_F0 = r_A & 0x20;
	//bank and AC omitted
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R0_A(void) {
        r_R0 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R1_A(void) {
        r_R1 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R2_A(void) {
        r_R2 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R3_A(void) {
        r_R3 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R4_A(void) {
        r_R4 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R5_A(void) {
        r_R5 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R6_A(void) {
        r_R6 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R7_A(void) {
        r_R7 = r_A;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_MOV_R0_IMM(void) {
	++r_PC;
	r_R0 = GetInstrMem(r_PC);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_MOV_R1_IMM(void) {
        ++r_PC;
        r_R1 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_R2_IMM(void) {
        ++r_PC;
        r_R2 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_R3_IMM(void) {
        ++r_PC;
        r_R3 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_R4_IMM(void) {
        ++r_PC;
        r_R4 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_R5_IMM(void) {
        ++r_PC;
        r_R5 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_R6_IMM(void) {
        ++r_PC;
        r_R6 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_R7_IMM(void) {
        ++r_PC;
        r_R7 = GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOV_T_A(void) {
#ifdef TESTSTATE
		if(r_PC == 0x28c)
			printf("r: %2x\n", r_A);
#endif
	r_T = r_A;
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_MOVD_A_P4(void) {
	throw;
        printf("%04x\tMOVD\tA,P4\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_A_P5(void) {
	throw;
        printf("%04x\tMOVD\tA,P5\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_A_P6(void) {
	throw;
        printf("%04x\tMOVD\tA,P6\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_A_P7(void) {
	throw;
        printf("%04x\tMOVD\tA,P7\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_P4_A(void) {
	throw;
        printf("%04x\tMOVD\tP4,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_P5_A(void) {
	throw;
        printf("%04x\tMOVD\tP5,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_P6_A(void) {
	throw;
        printf("%04x\tMOVD\tP6,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVD_P7_A(void) {
	throw;
        printf("%04x\tMOVD\tP7,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_MOVP_A_AT_A(void) {
        ++r_PC; //page "feature"
        r_A = GetInstrMem((r_PC & 0xf00) + r_A);
        m_totalCycles += 2;
};

void CPU8040::I_MOVP3_A_AT_A(void) {
        r_A = GetInstrMem(0x300 + r_A);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOVX_A_AT_R0(void) {
        r_A = m_WD1772->ByteFromWD1772();
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOVX_A_AT_R1(void) {
        r_A = m_WD1772->ByteFromWD1772();
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_MOVX_AT_R0_A(void) {
#ifdef TESTSTATE
	printf("PC: %4x\n", r_PC);
#endif
	m_WD1772->ByteToWD1772(r_A);
	++r_PC;
	m_totalCycles += 2;
};

void CPU8040::I_MOVX_AT_R1_A(void) {
#ifdef TESTSTATE
	printf("PC: %4x\n", r_PC);
#endif
        m_WD1772->ByteToWD1772(r_A);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_NOP(void) {
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_AT_R0(void) {
        r_A |= m_internalRAM[r_R0];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_AT_R1(void) {
        r_A |= m_internalRAM[r_R1];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_IMM(void) {
        ++r_PC;
        r_A  |= GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_ORL_A_R0(void) {
        r_A |= r_R0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R1(void) {
        r_A |= r_R1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R2(void) {
        r_A |= r_R2;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R3(void) {
        r_A |= r_R3;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R4(void) {
        r_A |= r_R4;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R5(void) {
        r_A |= r_R5;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R6(void) {
        r_A |= r_R6;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_A_R7(void) {
        r_A |= r_R7;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_ORL_BUS_IMM(void) {
	unsigned short int a = 0;
	a = GetInstrMem(r_PC + 1);
#ifdef TESTSTATE
	printf("%04x\tORL\tBUS,#%02x\n", r_PC, a);
#endif
	r_PC = r_PC + 2;
};

void CPU8040::I_ORL_P1_IMM(void) {
        ++r_PC;
        unsigned char tmp = GetInstrMem(r_PC);
        tmp |= r_Port1;
        NotifyChangePort1(r_Port1, tmp);
        r_Port1 = tmp;
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_ORL_P2_IMM(void) {
        ++r_PC;
        r_Port2 = ((GetInstrMem(r_PC) | r_Port2) & 0x3f) | (m_DriveNumber << 6);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_ORLD_P4_A(void) {
	throw;
        printf("%04x\tORLD\tP4,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_ORLD_P5_A(void) {
	throw;
        printf("%04x\tORLD\tP5,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_ORLD_P6_A(void) {
	throw;
        printf("%04x\tORLD\tP6,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_ORLD_P7_A(void) {
	throw;
        printf("%04x\tORLD\tP7,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_OUTL_BUS_A(void) {
	throw;
        printf("%04x\tOUTL\tBUS,A\n", r_PC);
        ++r_PC;
};

void CPU8040::I_OUTL_P1_A(void) {
        unsigned char tmp = r_A;
        NotifyChangePort1(r_Port1, tmp);
        r_Port1 = tmp;
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_OUTL_P2_A(void) {
        r_Port2 = (r_A & 0x3f) | (m_DriveNumber << 6);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_RET(void) {
	r_sp -= 2;
	r_sp &= 0xf;
	r_PC = ((m_internalStack[r_sp + 1] & 0xf) << 8) + (m_internalStack[r_sp ]);
        m_totalCycles += 2;
};

void CPU8040::I_RETR(void) {
        r_sp -= 2;
	r_sp &= 0xf;
	r_C = m_internalStack[r_sp + 1] & 0x80;
	r_F0 = m_internalStack[r_sp + 1] & 0x20;
        r_PC = ((m_internalStack[r_sp + 1] & 0xf) << 8) + (m_internalStack[r_sp]);
	m_IntPossible = true;
        m_totalCycles += 2;
};

void CPU8040::I_RL_A(void) {
	bool tmp = r_A & 0x80;
	r_A = r_A << 1;
	r_A = r_A | (tmp ? 0x01 : 0);
	++r_PC;
	++m_totalCycles;
};

void CPU8040::I_RLC_A(void) {
        bool tmp = r_C;
	r_C = r_A & 0x80;
        r_A = r_A << 1;
        r_A = r_A | (tmp ? 0x01 : 0);
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_RR_A(void) {
        bool tmp = r_A & 0x01;
        r_A = r_A >> 1;
        r_A = r_A | (tmp ? 0x80 : 0);
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_RRC_A(void) {
	bool tmp = r_C;
        r_C = r_A & 0x01;
        r_A = r_A >> 1;
        r_A = r_A | (tmp ? 0x80 : 0);
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_SEL_MB0(void) {
	r_dbf = false;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_SEL_MB1(void) {
        r_dbf = true;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_SEL_RB0(void) {
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_SEL_RB1(void) {
	throw;
        printf("%04x\tSEL\tRB1\n", r_PC);
        ++r_PC;
};

void CPU8040::I_STOP_TCNT(void) {
	m_TRunning = false;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_STRT_CNT(void) {
	throw;
        printf("%04x\tSTRT\tCNT\n", r_PC);
        ++r_PC;
};

void CPU8040::I_STRT_T(void) {
        m_TRunning = true;
	m_SpecialEventTable[CPUEVENTTIMERINCREASE] = m_totalCycles + 480;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_SWAP_A(void) {
	unsigned char tmp = r_A << 4;
	r_A = (r_A >> 4) + tmp;
	++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_UNDEF(void) {
	throw;
        unsigned short int a = 0;
        a = GetInstrMem(r_PC);
        printf("%04x\tUNDEF\t#%02x\n", r_PC, a);
        ++r_PC;
};

void CPU8040::I_XCH_A_AT_R0(void) {
	unsigned char tmp = r_A;
        r_A = m_internalRAM[r_R0];
	m_internalRAM[r_R0] = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_AT_R1(void) {
        unsigned char tmp = r_A;
        r_A = m_internalRAM[r_R1];
        m_internalRAM[r_R1] = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R0(void) {
        unsigned char tmp = r_A;
        r_A = r_R0;
        r_R0 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R1(void) {
#ifdef TESTSTATE
		if(r_PC == 0x27d)
			printf("r: %2x\n", r_A);
#endif
        unsigned char tmp = r_A;
        r_A = r_R1;
        r_R1 = tmp;
		m_internalRAM[1] = r_R1; //RAM duplication (omitted in the rest of the cpu emu)
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R2(void) {
        unsigned char tmp = r_A;
        r_A = r_R2;
        r_R2 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R3(void) {
        unsigned char tmp = r_A;
        r_A = r_R3;
        r_R3 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R4(void) {
        unsigned char tmp = r_A;
        r_A = r_R4;
        r_R4 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R5(void) {
        unsigned char tmp = r_A;
        r_A = r_R5;
        r_R5 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R6(void) {
        unsigned char tmp = r_A;
        r_A = r_R6;
        r_R6 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCH_A_R7(void) {
        unsigned char tmp = r_A;
        r_A = r_R7;
        r_R7 = tmp;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XCHD_A_AT_R0(void) {
	throw;
	printf("%04x\tXCHD\tA,@R0\n", r_PC);
	++r_PC;
};

void CPU8040::I_XCHD_A_AT_R1(void) {
	throw;
	printf("%04x\tXCHD\tA,@R1\n", r_PC);
	++r_PC;
};

void CPU8040::I_XRL_A_AT_R0(void) {
        r_A ^= m_internalRAM[r_R0];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_AT_R1(void) {
        r_A ^= m_internalRAM[r_R1];
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_IMM(void) {
        ++r_PC;
        r_A  ^= GetInstrMem(r_PC);
        ++r_PC;
        m_totalCycles += 2;
};

void CPU8040::I_XRL_A_R0(void) {
        r_A ^= r_R0;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R1(void) {
        r_A ^= r_R1;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R2(void) {
        r_A ^= r_R2;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R3(void) {
        r_A ^= r_R3;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R4(void) {
        r_A ^= r_R4;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R5(void) {
        r_A ^= r_R5;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R6(void) {
        r_A ^= r_R6;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::I_XRL_A_R7(void) {
        r_A ^= r_R7;
        ++r_PC;
        ++m_totalCycles;
};

void CPU8040::UpdatePSW(void) {
	r_psw = (r_C ? 0x80 : 0) + 0x0 + (r_F0 ? 0x20 : 0) + 0x0 + 0x08 + ((r_sp / 2) & 0x7);
};

//set a value in data memory
//void CPU8040::SetDataMem(unsigned short addr, unsigned char value);
//get a value from data memory
//unsigned char CPU8040::GetDataMem(unsigned short addr);

//set instruction Mem Pointer
void CPU8040::SetInstrMemPtr(void* ptr) {
	m_instrMemPtr = ptr;
};

//set ROM length
void CPU8040::SetROMLen(unsigned long myLen) {
	m_ROMLen = myLen;
};

//set WD1772 controller
void CPU8040::SetWD1772(WD1772Base* myWD1772) {
	m_WD1772 = myWD1772;
};

//set Serial controlling part
void CPU8040::SetSerial(SerialBase* mySerial) {
	m_Serial = mySerial;
};

//set instruction pointer
void CPU8040::SetInstrPtr(unsigned short myPtr) {
	r_PC = myPtr;
};

//get instruction pointer
unsigned short CPU8040::GetInstrPtr() {
	return r_PC;
};

//get a value from instruction memory
unsigned char CPU8040::GetInstrMem(unsigned short addr) {
	return *( (unsigned char *) m_instrMemPtr + addr);
};

//set a value in instruction memory
//for testing only
void CPU8040::SetInstrMem(unsigned short addr, unsigned char value){
        *( (unsigned char *) m_instrMemPtr + addr) = value;
};

//set drive number (0-3)
void CPU8040::SetDriveNumber(unsigned char DriveNo){
	if(DriveNo <= 3)
		m_DriveNumber = DriveNo;
};

//notify port 1 on changing bits
void CPU8040::NotifyChangePort1(unsigned char oldPort, unsigned char newPort){
	unsigned char hasChanged = oldPort ^ newPort;

	if(hasChanged & 0x03){
		m_WD1772->ChooseRegister(newPort & 0x03);
	};
	if(hasChanged & 0x08){
		m_WD1772->ChooseDoubleDensity(!(newPort & 0x08));
	};
	if(hasChanged & 0x10){
		if(!(newPort & 0x10)){
			m_WD1772->MasterReset();
		};
	};
	if(hasChanged & 0x20){
		m_WD1772->ChooseReadRegisters(newPort & 0x20);
	};
	if(hasChanged & 0x40){
		m_WD1772->ChooseSide(newPort & 0x40);
	};
	//if(hasChanged & 0x80){
		m_Serial->SerialOutBit(newPort & 0x80);
	//};
};

//when CPU is reset
void CPU8040::ResetCPU(){
	if(m_instrMemPtr == NULL)
		throw;
	if(m_WD1772 == NULL)
		throw;
	if(m_Serial == NULL)
		throw;
	m_totalCycles = 0;
	m_delayCycles = 0;
	r_PC = 0;
	r_sp = 0;
	r_dbf = false;
	r_F0 = false;
	r_F1 = false;
	r_TF = false;
	m_ExternalIntEnabled = false;
	m_TCNTIntEnabled = false;
	m_IntPossible = true;
	m_TRunning = false;
	for (int loop = 0;loop < SPECIALEVENTCOUNT; ++loop){
		m_SpecialEventTable[loop] = MAXCYCLE;
	}
	DetermineNextEventCount();
	r_Port1 = 0xff;
	r_Port2 = 0x3f;

	//TODO: Rest TBD
	//reset WD1772
};

/*
vim:ts=4:sw=4:
*/
