// interface functions
// interface.c

#include <stdio.h>
#include <string.h>
#include "interface.h"
#include "floppy.h"
#ifdef WITH_CAPS
#include "xatremu_caps.h"
#endif

using namespace xatremu;

#define XATRMAXDRIVE 4
xatremu::FloppyDrive * XATREMU_FDrive[XATRMAXDRIVE];
char * XATREMU_XF551ROMName;
unsigned char XATREMU_OutByteDrive;

extern "C" char XATREMU_Init(){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		XATREMU_FDrive[loop] = NULL;
	};
	XATREMU_XF551ROMName = NULL;
	XATREMU_OutByteDrive = 0;

#ifdef WITH_CAPS
	CAPSInit();
#endif
};

extern "C" void XATREMU_Free(){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		delete XATREMU_FDrive[loop];
		XATREMU_FDrive[loop] = NULL;
	};
	if(XATREMU_XF551ROMName != NULL){
		delete[] XATREMU_XF551ROMName;
		XATREMU_XF551ROMName = NULL;
	};

#ifdef WITH_CAPS
	CAPSExit();
#endif
};

extern "C" void XATREMU_SetXF551ROMName(char * ROMName){
	if(XATREMU_XF551ROMName != NULL){
		delete[] XATREMU_XF551ROMName;
		XATREMU_XF551ROMName = NULL;
	};
	if(ROMName != NULL){
		XATREMU_XF551ROMName = new char[strlen(ROMName)+5];
		strncpy(XATREMU_XF551ROMName, ROMName, strlen(ROMName)+4);
	};
};

extern "C" void XATREMU_SetDrive(unsigned char DriveNo, unsigned char DriveFlavour){
	if(DriveFlavour == XATREMU_DRIVEFLAVOUROFF){
		if(XATREMU_FDrive[DriveNo] != NULL){
			delete XATREMU_FDrive[DriveNo];
			XATREMU_FDrive[DriveNo] = NULL;
		};
	};
	if((DriveFlavour == XATREMU_DRIVEFLAVOURXF551) || (DriveFlavour == XATREMU_DRIVEFLAVOURXF551_288RPM)){
		if(XATREMU_FDrive[DriveNo] != NULL){
			delete XATREMU_FDrive[DriveNo];
			XATREMU_FDrive[DriveNo] = NULL;
		};
		XATREMU_FDrive[DriveNo] = new FloppyDrive(XATREMU_XF551ROMName);
		XATREMU_FDrive[DriveNo]->SetDriveNumber(DriveNo);
		XATREMU_FDrive[DriveNo]->SetTo288RPM(DriveFlavour == XATREMU_DRIVEFLAVOURXF551_288RPM ? true : false);
	};
};

extern "C" unsigned char XATREMU_SetImage(unsigned char DriveNo, char * ImageName){
	if(XATREMU_FDrive[DriveNo] != NULL){
		return XATREMU_FDrive[DriveNo]->SetImage(ImageName) ? 1 : 0;
	};
	return 0;
};

extern "C" unsigned char XATREMU_EjectImage(unsigned char DriveNo){
	if(XATREMU_FDrive[DriveNo] != NULL){
		XATREMU_FDrive[DriveNo]->EjectImage();
	};
};

//Sets interrupt line (to low when SKCTL & 0x70 == 0x20)
extern "C" void XATREMU_SetIntLineAtari(unsigned char IntLine){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			XATREMU_FDrive[loop]->SetIntLineAtari(IntLine ? true : false);
		};
	};
};

extern "C" void XATREMU_SerialByteFromAtari(unsigned char AtariByte){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			XATREMU_FDrive[loop]->SerialByteFromAtari(AtariByte);
		};
	};
};

extern "C" unsigned char XATREMU_ReadyForNextByteFromAtari(){
	unsigned char ReadyStatus = 1;
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			if(!(XATREMU_FDrive[loop]->ReadyForNextInByte())){
				ReadyStatus = 0;
			};
		};
	};
	return ReadyStatus;
};

extern "C" unsigned char XATREMU_SerialByteToAtari(){
	if(XATREMU_FDrive[XATREMU_OutByteDrive] != NULL){
		return XATREMU_FDrive[XATREMU_OutByteDrive]->SerialByteToAtari();
	};
};

extern "C" unsigned char XATREMU_SerialByteForAtariAvailable(){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			if(XATREMU_FDrive[loop]->SerialByteForAtariAvailable()){
				XATREMU_OutByteDrive = loop;
				return 1;
			};
		};
	};
	return 0;
};

extern "C" void XATREMU_RunFloppiesOneScanline(){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			XATREMU_FDrive[loop]->RunCPUOneScanline();
		};
	};
};

extern "C" void XATREMU_RunFloppiesForCycles(unsigned int numOfCycles){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			XATREMU_FDrive[loop]->RunCPUForCycles(numOfCycles);
		};
	};
};

extern "C" void XATREMU_SIOSpeed(unsigned char Speed){
	for(int loop = 0; loop < XATRMAXDRIVE; ++loop){
		if(XATREMU_FDrive[loop] != NULL){
			XATREMU_FDrive[loop]->SetSIOSpeed(Speed);
		};
	};
};

extern "C" void XATREMU_SetWriteProtect(unsigned char DriveNo, unsigned char WriteProtected){
	if(XATREMU_FDrive[DriveNo] != NULL){
		XATREMU_FDrive[DriveNo]->SetWriteProtect(WriteProtected);
	};
};

extern "C" unsigned char XATREMU_FloppyState(unsigned char DriveNo){
	if(XATREMU_FDrive[DriveNo] != NULL){
		return XATREMU_FDrive[DriveNo]->FloppyState();
	};
	return 0;
};

